/* Copyright (c) 1994-1996 David Hogan, see README for licence details */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "autoconfig.h"
#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif
#include "dat.h"
#include "fns.h"
#include "config.h"

Client	*clients = NULL;
Client	*current = NULL;
int cycling = 0;

void setactive (Client *c, int on) {
	if (c->parent == c->screen->root) {
		fprintf(stderr, WMNAME ": bad parent in setactive; dumping core\n");
		return;	/* ### BTA 21032006 */
		/* abort(); */
	}
	if (on) {
		XUngrabButton(dpy, AnyButton, AnyModifier, c->parent);
		XGrabButton(dpy, AnyButton, Mod1Mask, c->parent, False,
			ButtonMask, GrabModeAsync, GrabModeAsync, None, None);
// Second Grab from Sync to Async. Xmas 2000, bta.
		grabkeys(c);
		XSetInputFocus(dpy, c->window, RevertToPointerRoot, timestamp());
		if (c->proto & Ptakefocus)
			sendcmessage(c->window, wm_protocols, wm_take_focus, 0);
		cmapfocus(c);
	} else
		XGrabButton(dpy, AnyButton, AnyModifier, c->parent, False,
			ButtonMask, GrabModeAsync, GrabModeAsync, None, None);
// Second Grab from Sync to Async. Xmas 2000, bta.
	draw_border(c, on);
}

void draw_border (Client *c, int active) {
	XSetWindowBackground(dpy, c->parent, active ? c->screen->activeborder : c->screen->inactiveborder);
	XClearWindow(dpy, c->parent);
	if (c->hold && active)
		XDrawRectangle(dpy, c->parent, c->screen->gc,
				INSET, INSET, c->dx+BORDER-INSET, c->dy+BORDER-INSET);
}

void active (Client *c) {
	Client *cc;

	if (c == 0) {
		fprintf(stderr, WMNAME ": active(c==0)\n");
		return;
	}
	if (c == current)
		return;
	if (current) {
		setactive(current, 0);
		if (current->screen != c->screen)
			cmapnofocus(current->screen);
	}
	setactive(c, 1);
	for (cc = clients; cc; cc = cc->next)
		if (cc->revert == c)
			cc->revert = c->revert;
	c->revert = current;
	while (c->revert && !normal(c->revert))
		c->revert = c->revert->revert;
	current = c;
}

void nofocus (void) {
	static Window w = 0;
	int mask;
	XSetWindowAttributes attr;
	Client *c;

	if (current) {
		setactive(current, 0);
		for (c = current->revert; c; c = c->revert)
			if (normal(c)) {
				active(c);
				return;
			}
		cmapnofocus(current->screen);
		/* if no candidates to revert to, fall through */
	}
	current = 0;
	if (w == 0) {
		mask = CWOverrideRedirect;
		attr.override_redirect = 1;
		w = XCreateWindow(dpy, screens[0].root, 0, 0, 1, 1, 0,
			CopyFromParent, InputOnly, CopyFromParent, mask, &attr);
		XMapWindow(dpy, w);
	}
	XSetInputFocus(dpy, w, RevertToPointerRoot, timestamp());
}

void top (Client *c) {
	Client **l, *cc;

	l = &clients;
	for (cc = *l; cc; cc = *l) {
		if (cc == c) {
			*l = c->next;
			c->next = clients;
			clients = c;
			return;
		}
		l = &cc->next;
	}
	fprintf(stderr, WMNAME ": %x not on client list in top()\n", (unsigned)c);
}

void cycle (int raise) {
	Client *c = current->next, *oldcur = current, *n;
	int xit = 0;
	XEvent ev;

	XGrabKeyboard(dpy, current->parent, False, GrabModeAsync, GrabModeAsync, CurrentTime);
	cycling = 1;
	do {
		Client *oc = c;
		while (c && (!normal(c) || (c->parent == c->screen->root))) {
			c = c->next;
			if (!c) c = clients;
			if (c == oc) break;  /* just in case */
		}
		if (!c) break;
		if (raise) XRaiseWindow(dpy, c->parent);
		n = c->next;
		active(c);
		for (;;) {
			XMaskEvent(dpy, KeyPressMask|KeyReleaseMask, &ev);
			if ((ev.type == KeyRelease &&
					ev.xkey.keycode == XKeysymToKeycode(dpy,XK_Alt_L)) ||
					(ev.xkey.state&CYCLE_MOD) != CYCLE_MOD) {
				xit = 1;
				break;
			}
			if (ev.type == KeyPress)
				if (ev.xkey.keycode == XKeysymToKeycode(dpy,CYCLE_KEY)) {
					c = n;
					break;
				}
		}
	} while (!xit);
	for (;;)
		if (!XCheckMaskEvent(dpy, KeyPressMask|KeyReleaseMask, &ev))
			break;
	XUngrabKeyboard(dpy, CurrentTime);
	cycling = 0;
	if (c) {
		top(oldcur);
		top(c);
		active(c);  /* now the keys will be grabbed */
	}
}

Client *getclient (Window w, int create) {
	Client *c;

	if (w == 0 || getscreen(w))
		return 0;

	for (c = clients; c; c = c->next)
		if (c->window == w || c->parent == w)
			return c;

	if (!create)
		return 0;

	c = (Client *)malloc(sizeof(Client));
	memset(c, 0, sizeof(Client));
	c->window = w;
	/* c->parent will be set by the caller */
	c->parent = None;
	c->reparenting = 0;
	c->state = WithdrawnState;
	c->init = 0;
	c->cmap = None;
	c->label = c->class = 0;
	c->revert = 0;
	c->hold = 0;
	c->ncmapwins = 0;
	c->cmapwins = 0;
	c->wmcmaps = 0;
	c->next = clients;
	clients = c;
	return c;
}

void rmclient (Client *c) {
	Client *cc;

	for (cc = current; cc && cc->revert; cc = cc->revert)
		if (cc->revert == c)
			cc->revert = cc->revert->revert;

	if (c == clients)
		clients = c->next;
	for (cc = clients; cc && cc->next; cc = cc->next)
		if (cc->next == c)
			cc->next = cc->next->next;

	if (hidden(c))
		unhidec(c, 0);

	if( workspace_out(c) )
		workspace_delete(c);

	if (c->parent != c->screen->root)
		XDestroyWindow(dpy, c->parent);

	c->parent = c->window = None;		/* paranoia */
	if (current == c) {
		current = c->revert;
		if (current == 0)
			nofocus();
		else {
			if (current->screen != c->screen)
				cmapnofocus(c->screen);
			setactive(current, 1);
		}
	}
	if (c->ncmapwins != 0) {
		XFree((char *)c->cmapwins);
		free((char *)c->wmcmaps);
	}
	if (c->iconname != 0)
		XFree((char*) c->iconname);
	if (c->name != 0)
		XFree((char*) c->name);
	if (c->instance != 0)
		XFree((char*) c->instance);
	if (c->class != 0)
		XFree((char*) c->class);
	memset(c, 0, sizeof(Client));		/* paranoia */
	free(c);
}

#ifdef SHAPE
void set_shape(Client *c) {
   int n, order;
   XRectangle *rect;

   if (!have_shape) {
       return;
   }
   rect = XShapeGetRectangles(dpy, c->window, ShapeBounding, &n, &order);
   if (n > 1) {
       XShapeCombineShape(dpy, c->parent, ShapeBounding, c->border,
               c->border, c->window, ShapeBounding, ShapeSet);
   }
   XFree((void*)rect);
}
#endif
