#include <stdio.h>
#include <signal.h>
#include <X11/X.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include "dat.h"
#include "fns.h"
#include "config.h"

extern int cycling;

void initkeys (void) {
	int i;
	XGrabKey(dpy, XKeysymToKeycode(dpy,XK_Return), HOTKEY_MOD, screens[0].root, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy,EXIT_KEY), HOTKEY_MOD, screens[0].root, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy,NETSCAPE_KEY), HOTKEY_MOD, screens[0].root, True, GrabModeAsync, GrabModeAsync);
	for( i = 0; i < MAX_WORKSPACE; i++ )
		XGrabKey(dpy, XKeysymToKeycode(dpy,WORKSPACE_KEY+i), HOTKEY_MOD, screens[0].root, True, GrabModeAsync, GrabModeAsync);
	/* It's difficult to customize the following keys, plus I want them as they are now :-) */
	XGrabKey(dpy, XKeysymToKeycode(dpy,XK_Left), (ControlMask), screens[0].root, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy,XK_Right), (ControlMask), screens[0].root, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy,XK_bracketleft), HOTKEY_MOD, screens[0].root, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy,XK_bracketright), HOTKEY_MOD, screens[0].root, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy,XK_h), HOTKEY_MOD|ShiftMask, screens[0].root, True, GrabModeAsync, GrabModeAsync);
}

void key (XKeyEvent *e) {
	Client *c;
	ScreenInfo *s;
	KeySym key;

	s = getscreen(e->root);
	if (s == 0)
		return;

	curtime = e->time;
	if (getscreen(e->root) == 0)
		return;
	c = getclient(e->window, 0);

	key = XKeycodeToKeysym(dpy,e->keycode,0);
	switch ( key ) {
	case XK_Up:
	case XK_KP_Up:
		move(c, 0, 1, 0, -16); break;
	case XK_Down:
	case XK_KP_Down:
		move(c, 0, 1, 0, 16); break;
	case XK_Left:
	case XK_KP_Left:
		if( (e->state & Mod1Mask) == Mod1Mask ) {
			move(c, 0, 1, -16, 0);
		} else {
			prev_workspace();
		} 
		break;
	case XK_Right:
	case XK_KP_Right:
		if( (e->state & Mod1Mask) == Mod1Mask ) {
			move(c, 0, 1, 16, 0);
		} else {
			next_workspace();
		}
		break;
    /* redefined to position windows
	 * case XK_KP_Insert:
	 *	resize(c, 0); break; */
	/* not much use for this, really...
	 * case XK_KP_Delete:
	 * 	move(c, 0, 1, 0, 0); break; */
	case XK_Insert:
	case XK_KP_Insert:
		move(c, 0, 0, BORDER, BORDER); break;
	case XK_Delete:
	case XK_KP_Delete:
		move(c, 0, 0, BORDER, DisplayHeight(dpy, c->screen->num)-c->dy-BORDER); break;
	case XK_Home:
	case XK_KP_Home:
		move(c, 0, 0, (DisplayWidth(dpy, c->screen->num)-c->dx-BORDER)/2, BORDER); break;
	case XK_End:
	case XK_KP_End:
		move(c, 0, 0, (DisplayWidth(dpy, c->screen->num)-c->dx-BORDER)/2, DisplayHeight(dpy, c->screen->num)-c->dy-BORDER); break;
	case XK_Page_Up:
	case XK_KP_Page_Up:					 /* +1		-1 */
		move(c, 0, 0, DisplayWidth(dpy, c->screen->num)-c->dx-BORDER, BORDER); break;
	case XK_Page_Down:
	case XK_KP_Page_Down:
		move(c, 0, 0, DisplayWidth(dpy, c->screen->num)-c->dx-BORDER, DisplayHeight(dpy, c->screen->num)-c->dy-BORDER); break;
	case XK_Return:
		spawn(s); break;
	case CYCLE_KEY:
		cycle(!((e->state&ControlMask)==ControlMask)); break;
	case XK_Escape:
		delete(c, (e->state&ShiftMask)==ShiftMask); break;
	case XK_h:
		if( (e->state & ShiftMask) == ShiftMask ) {
			if( numhidden > 0 ) {
				unhide( 0, 1 );
			} 
		} else {
			hide(c);
		}
		break;
	case XK_m: {
		int w = DisplayWidth( dpy, c->screen->num );
		int h = DisplayHeight( dpy, c->screen->num );
		//XLowerWindow( dpy, c->parent );
		//XRaiseWindow( dpy, c->parent );
		XMoveResizeWindow( dpy, c->parent, 0, 0, w, h );
/* I had modified it trying to get that border drawn also on the bottom side,
 * but without luck. */
/*		XMoveResizeWindow( dpy, c->window, BORDER-1, BORDER-1,
			w-BORDER-1, h-BORDER-1 );	*/
		c->x = c->y = BORDER;
		c->dx = w-((BORDER)*2);
		c->dy = h-((BORDER)*2);
		XMoveResizeWindow( dpy, c->window, c->x, c->y, c->dx, c->dy );
		sendconfig( c );	/* Do this do any good ? */ 
		break;
	}
	case XK_w: {
		if( (e->state & ShiftMask) == ShiftMask )
			c->dx -= 16;
		else
			c->dx += 16;
		if( c->dx < WM_WINDOW_SAFE_DX )
			c->dx = WM_WINDOW_SAFE_DX;
		XMoveResizeWindow( dpy, c->parent, c->x-BORDER, c->y-BORDER,
			c->dx+2*(BORDER-1), c->dy+2*(BORDER-1) );
		// sendconfig( c );	// Why it doesn't work ? [a clueless bta]
		XMoveResizeWindow( dpy, c->window, BORDER-1, BORDER-1, 
			c->dx, c->dy );
		break;
	}
	case XK_e: {
		if( (e->state & ShiftMask) == ShiftMask )
			c->dy -= 16;
		else
			c->dy += 16;
		if( c->dy < WM_WINDOW_SAFE_DY )
			c->dy = WM_WINDOW_SAFE_DY;
		XMoveResizeWindow( dpy, c->parent, c->x-BORDER, c->y-BORDER,
			 c->dx+2*(BORDER-1), c->dy+2*(BORDER-1) );
		// sendconfig( c );
		XMoveResizeWindow( dpy, c->window, BORDER-1, BORDER-1, 
			c->dx, c->dy );
		break;
	}
	case XK_bracketleft:
		workspace_less(); break;
	case XK_bracketright:
		workspace_more(); break;
	case EXIT_KEY:
		raise( SIGTERM ); break;
	case NETSCAPE_KEY:
		spawn_netscape( s ); break;
	default:
		if( ( key >= WORKSPACE_KEY )&&( key < (WORKSPACE_KEY + MAX_WORKSPACE) ) ) {
			set_workspace( key - WORKSPACE_KEY );
		}
	}
}

void grabkeys (Client *c) {
	if (cycling)
		return;
	XUngrabKeyboard(dpy, CurrentTime);
	XGrabKey(dpy, AnyKey, HOTKEY_MOD, c->parent, True, GrabModeAsync, GrabModeAsync);
	XGrabKey(dpy, XKeysymToKeycode(dpy, CYCLE_KEY), CYCLE_MOD, c->parent, True, GrabModeAsync, GrabModeAsync);
}
