/* Copyright (c) 1994-1996 David Hogan, see README for licence details */
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "dat.h"
#include "fns.h"
#include "config.h"

Client	*hiddenc[MAXHIDDEN];

int	numhidden;

char	*b3items[B3FIXED+MAXHIDDEN+1] = {
	"xterm",
	"Reshape",
	"Delete",
	"Hide",
	"PopUp",
	"Restart",
        "C-A-x to Exit",
	0,
};

Menu	b3menu = {
	b3items,
};

Menu	egg = {
	version,
};

void button (XButtonEvent *e) {
	int n, shift;
	Client *c;
	Window dw;
	ScreenInfo *s;

	curtime = e->time;
	s = getscreen(e->root);
	if (s == 0)
		return;
	c = getclient(e->window, 0);
	if (c) {
		e->x += c->x - BORDER + 1;
		e->y += c->y - BORDER + 1;
	} else if (e->window != e->root)
		XTranslateCoordinates(dpy, e->window, s->root, e->x, e->y,
				&e->x, &e->y, &dw);
	switch (e->button) {
	case MOUSE_BUTTON1:
		move(c, MOUSE_BUTTON1, 0, 0, 0);
		return;
	case MOUSE_BUTTON2:
		menuhit(e, &egg);
		return;
	default:
		return;
	case MOUSE_BUTTON3:
		if (c) {
			resize(c, MOUSE_BUTTON3);
			return;
		}
	}

	if (current && current->screen == s)
		cmapnofocus(s);
	switch (n = menuhit(e, &b3menu)) {
	case 0: 	/* New */
		spawn(s);
		break;
	case 1: 	/* Reshape */
		reshape(selectwin(1, 0, s), MOUSE_BUTTON3);
		break;
	case 2: 	/* Delete */
		shift = 0;
		c = selectwin(1, &shift, s);
		delete(c, shift);
		break;
	case 3: 	/* Hide */
		hide(selectwin(1, 0, s));
		break;
        case 4:         /* PopUp */
	    {
		Client *mc;
		for( mc = clients; mc; mc = mc->next ) {
			if(( mc != c )&&( workspace_mapped( mc ) < 0 )
			     &&(mc->parent != e->root )) {
fprintf(stderr, WMNAME ": PopUp on Client %08X (Parent %08X, Root: %08X)\n",
	mc->window, mc->parent, e->root );
				workspace_unhide( mc );
			}
		}
	    }
	    break;
        case 5:         /* Restart */
                fprintf(stderr, "*** " WMNAME " restarting ***\n");
                cleanup();
                execvp(myargv[0], myargv);
                perror(WMNAME ": exec failed");
		break;
        case 6:         /* Exit */
		// Too easy to exit the wm without a real intention...
		// Use CTRL+ALT+X instead :-)
                //raise( SIGTERM );
                break;
	default:	/* unhide window */
		unhide(n - B3FIXED, 1);
		break;
	case -1:	/* nothing */
		break;
	}
	if (current && current->screen == s)
		cmapfocus(current);
}

void spawn_bta( ScreenInfo *s, void (*exec_func)(void) ) {
	/*
	 * ugly dance to avoid leaving zombies.  Could use SIGCHLD,
	 * but it's not very portable.
	 */
	if (fork() == 0) {
		if (fork() == 0) {
			close(ConnectionNumber(dpy));
			if (s->display[0] != '\0')
				putenv(s->display);
			if (termprog != NULL) {
				execl(shell, shell, "-c", termprog, 0);
				fprintf(stderr, WMNAME ": exec %s", shell);
				perror(" failed");
			}
//			execlp( "xterm", "xterm", XTERM_PARAMS, 0 );
			exec_func();
			perror(WMNAME ": exec xterm failed");
			exit(1);
		}
		exit(0);
	}
	wait((int *) 0);
}

static void exec_xterm( void ) {
	/* execlp( "rxvt", "rxvt", XTERM_PARAMS, 0 ); */
	execlp( "xterm", "xterm", XTERM_PARAMS, 0 );
}

static void exec_netscape( void ) {
/*	execlp( "netscape", "netscape", NETSCAPE_PARAMS, 0 ); */
/*	execlp( "opera", "opera", NETSCAPE_PARAMS, 0 ); */
	execlp( "firefox", "firefox", NETSCAPE_PARAMS, 0 );
}

void spawn(ScreenInfo *s) {
	spawn_bta( s, exec_xterm );
}

void spawn_netscape( ScreenInfo *s ) {
	spawn_bta( s, exec_netscape );
}

void reshape (Client *c, unsigned int b) {
	int odx, ody;

	if (c == 0)
		return;
	odx = c->dx;
	ody = c->dy;
	if (sweep(c, b) == 0)
		return;
	if( normal( c ) ) {
		active(c);
		top(c);
		XRaiseWindow(dpy, c->parent);
		XMoveResizeWindow(dpy, c->parent, c->x-BORDER, c->y-BORDER,
						c->dx+2*(BORDER-1), c->dy+2*(BORDER-1));
		if (c->dx == odx && c->dy == ody)
			sendconfig(c);
		else
			XMoveResizeWindow(dpy, c->window, BORDER-1, BORDER-1, c->dx, c->dy);
	}
}

void move (Client *c, unsigned int b, int rel, int dx, int dy) {
	if (c == 0)
		return;
	if (b) {
		if (drag(c, b) == 0)
			return;
	} else {
		if (rel) {
			c->x += dx; c->y += dy;
		} else {
			c->x = dx; c->y = dy;
		}
	}
	if( normal( c ) ) {
		active(c);
		top(c);
		XRaiseWindow(dpy, c->parent);
		XMoveWindow(dpy, c->parent, c->x-BORDER, c->y-BORDER);
		sendconfig(c);
	}
}

void resize (Client *c, unsigned int b) {
	if (c == 0)
		return;
	if( normal( c ) ) {
		XLowerWindow(dpy, c->parent);
		sendconfig(c);
	}
}

void delete (Client *c, int shift) {
	if (c == 0)
		return;
	if ((c->proto & Pdelete) && !shift)
		sendcmessage(c->window, wm_protocols, wm_delete, 0);
	else
		XKillClient(dpy, c->window);		/* let event clean up */
}

void hide (Client *c) {
	if (c == 0 || numhidden == MAXHIDDEN)
		return;
	if (hidden(c)) {
		fprintf(stderr, WMNAME ": already hidden: %s\n", c->label);
		return;
	}
	XUnmapWindow(dpy, c->parent);
	XUnmapWindow(dpy, c->window);
	mysetstate(c, IconicState);
	if (c == current)
		nofocus();
	hiddenc[numhidden] = c;
	b3items[B3FIXED+numhidden] = c->label;
	numhidden++;
	b3items[B3FIXED+numhidden] = 0;
}

void unhide (int n, int map) {
	Client *c;
	int i;

	if (n >= numhidden) {
		fprintf(stderr, WMNAME ": unhide: n %d numhidden %d\n", n, numhidden);
		return;
	}
	c = hiddenc[n];
	if (!hidden(c)) {
		fprintf(stderr, WMNAME ": unhide: not hidden: %s(0x%x)\n",
			c->label, (unsigned)c->window);
	} else {
        if (map) {
            XMapWindow(dpy, c->window);
            XMapRaised(dpy, c->parent);
            mysetstate(c, NormalState);
            active(c);
            top(c);
        }
    }

	numhidden--;
	for (i = n; i < numhidden; i ++) {
		hiddenc[i] = hiddenc[i+1];
		b3items[B3FIXED+i] = b3items[B3FIXED+i+1];
	}
	b3items[B3FIXED+numhidden] = 0;
}

void unhidec (Client *c, int map) {
	int i;

	for (i = 0; i < numhidden; i++)
		if (c == hiddenc[i]) {
			unhide(i, map);
			return;
		}
	/* make this quiet because some windows need to be unhidden when
	 * spawned (don't ask me why...)  --kja3 */
	/*fprintf(stderr, WMNAME ": unhidec: not hidden: %s(0x%x)\n",
		c->label, (unsigned)c->window);*/
}

void renamec (Client *c, char *name) {
	int i;

	if (name == 0)
		name = "???";
	c->label = name;
	if (!hidden(c))
		return;
	for (i = 0; i < numhidden; i++)
		if (c == hiddenc[i]) {
			b3items[B3FIXED+i] = name;
			return;
		}
}
