#include <stdlib.h>
#include <assert.h>

#include "xalloc.h"
#include "vector.h"

static void vector_realloc( VECTOR *v ) {
	if( v->size == v->max ) {
		v->max *= 2;
		v->data = xrealloc( v->data, v->max * sizeof( void * ) );
	}
}

VECTOR *vector_init( void ) {
	VECTOR *v = xmalloc( sizeof( VECTOR ) );
	v->size = 0;
	v->max = 10;
	v->data = xcalloc( v->max * sizeof( void * ) );

	/* Strange arch have NULL pointer != 0; at least stop here...
         If you have one of those machine, you should do the following things:
	   1) tell me what machine that is
	   2) put a loop here, and explicitly init all the pointers
	   3) put a similar loop in vector_realloc
         */ 
	assert( v->data[0] == NULL );
	return( v );
}

void vector_destroy( VECTOR *v, VECTOR_CALLBACK c_func, void *c_data ) {
	vector_iterate( v, c_func, c_data );
	xfree( v->data );
	xfree( v );
}

int vector_iterate( VECTOR *v, VECTOR_CALLBACK c_func, void *c_data ) {
	int i, r;

	if( !c_func ) {
		return( -1 );
	}

	for( i=0; i<v->size; i++ ) {
		r = c_func( v->data[i], c_data );
		if( r ) {
			return( r );
		}
	}
	return( 0 );
}

int vector_len( VECTOR *v ) {
	return( v->size );
}

int vector_find( VECTOR *v, void *d ) {
	int i;
	for( i=0; i<v->size; i++ ) {
		if( v->data[i] == d ) {
			return( i );
		}
	}
	return( -1 );
}

void vector_push( VECTOR *v, void *d ) {
	vector_realloc( v );
	v->data[ v->size++ ] = d;
}

void *vector_pop( VECTOR *v ) {
	if( v->size == 0 ) {
		return NULL;
	}
	return( v->data[ --v->size ] );
}

void *vector_at( VECTOR *v, int pos ) {
	if(( pos < 0 )||( pos >= v->size )) {
		return NULL;
	}
	return( v->data[ pos ] );
}

void vector_replace_at( VECTOR *v, int pos, void *data ) {
	if(( pos < 0 )||( pos >= v->size )) {
		return;
	}
	v->data[ pos ] = data;
}

void vector_del_at( VECTOR *v, int pos ) {
	int i;
	if(( pos < 0 )||( pos >= v->size )) {
		return;
	}
	for( i=pos; i<v->size; i++ ) {
		int nexti = i+1;
		if( nexti < v->max ) {
			v->data[i] = v->data[nexti];
		} else {
			break;
		}
	}
	v->data[i] = NULL;
    v->size--;
}

void vector_del( VECTOR *v, void *data ) {
	vector_del_at( v, vector_find( v, data ) );
}

int vector_cycle( VECTOR *v, int pos ) {
	if(( pos < 0 )||( pos >= v->size )) {
		return( 0 );
	}
	if( ++pos < v->size ) {
		return( pos );
	}
	return( 0 );
}
